"""Initializes the csv_io module of the datagrunt package."""

from datagrunt.core.csv_io.csvcomponents import (
    CSVDelimiter,
    CSVDialect,
    CSVColumns,
    CSVColumnNameNormalizer,
    CSVComponents,
    CSVRows,
    CSVStringSample
)
from datagrunt.core.csv_io.engines import (
    CSVEngineProperties,
    CSVReaderDuckDBEngine,
    CSVReaderPolarsEngine,
    CSVWriterDuckDBEngine,
    CSVWriterPolarsEngine
)
from datagrunt.core.csv_io.factories import CSVEngineFactory

__all__ = [
    'CSVDelimiter',
    'CSVDialect',
    'CSVColumns',
    'CSVColumnNameNormalizer',
    'CSVComponents',
    'CSVRows',
    'CSVStringSample',
    'CSVEngineProperties',
    'CSVReaderDuckDBEngine',
    'CSVReaderPolarsEngine',
    'CSVWriterDuckDBEngine',
    'CSVWriterPolarsEngine',
    'CSVEngineFactory'
]



"""Initializes the databases module of the datagrunt package."""

from datagrunt.core.databases.databases import DuckDBDatabase, DuckDBQueries

__all__ = [
    'DuckDBDatabase',
    'DuckDBQueries'
]



# Import key classes, functions, or submodules that should be available at the package level
from datagrunt.csv_api.csvreader import CSVReader
from datagrunt.csv_api.csvwriter import CSVWriter
from datagrunt.csv_api.csvai import CSVSchemaReportAIGenerated


# You can define __all__ to specify what gets imported with "from package import *"
__all__ = ['CSVReader', 'CSVWriter', 'CSVSchemaReportAIGenerated']



"""Factory module for creating AI factory instances."""

# local libraries
from datagrunt.core.ai.engines import AIEngineProperties, GoogleAIEngine

class AIEngineFactory:
    """Factory class for creating AI engine instances."""
    AI_ENGINES = {
                'google': GoogleAIEngine,
            }

    def __init__(self, api_key, engine, **kwargs):
        """
        Initialize the AI Engine Factory class.

        Args:
            api_key (str): API key for the AI provider.
            engine (str): type of engine to create by the factory.
            **kwargs: Additional parameters for the AI provider.
        """
        self.api_key = api_key
        self.engine = engine.lower().replace(' ', '')
        self.kwargs = kwargs
        if self.engine not in AIEngineProperties.valid_engines:
            raise ValueError(AIEngineProperties.value_error_message.format(engine=self.engine))

    def create_engine(self):
        """Create an AI engine instance."""
        engine_class = self.AI_ENGINES.get(self.engine)
        if engine_class:
            return engine_class(self.api_key, **self.kwargs)
        else:
            raise ValueError(f"Unsupported AI engine: {self.engine}")



"""Module for using AI tools to analyze CSV files."""

# standard library imports
import json

# local imports
from datagrunt.core import (
    AIEngineFactory,
    AIEngineProperties,
    CSVStringSample,
    prompts
)

class CSVSchemaReportAIGenerated:
    """Class to generate a CSV schema report using AI tools."""

    def __init__(self, filepath, engine, api_key=None, **kwargs):
        """Initialize the CSV Schema Report class."""
        self.filepath = filepath
        self.engine = engine.lower().replace(' ', '')
        self.api_key = api_key
        self.kwargs = kwargs
        if self.engine not in AIEngineProperties.valid_engines:
            raise ValueError(f"Unsupported AI engine: {self.engine}")
        if self.kwargs.pop('ground_google_search', False):
            raise ValueError("Grounding in Google Search is not supported for this class.")

    def _create_engine(self):
        """Create an AI engine instance."""
        return AIEngineFactory(self.api_key, self.engine, **self.kwargs).create_engine()

    def _get_ai_response(self, model, prompt, system_instructions):
        """Send a request to the AI engine and handle potential errors.

        Args:
            model (str): The name of the model to use.
            prompt (str): The prompt to send to the model.
            system_instructions (str): System instructions to guide the model's response.

        Returns:
            dict: The JSON response from the AI engine.

        Raises:
            ValueError: If the response is not valid JSON.
            TypeError: If the response is not a text string.
            RuntimeError: For any other unexpected errors.
        """
        try:
            response_text = self._create_engine().generate_content(
                model=model,
                prompt=prompt,
                system_instruction=system_instructions
            )
            return json.loads(response_text)
        except json.JSONDecodeError as e:
            error_message = (
                "The model's response was not a valid JSON object. "
                "This can happen if the `max_tokens` parameter is too low, cutting off the response. "
                f"Details: {e}\n"
                f"Model Response: {response_text[:500]}..."
            )
            raise ValueError(error_message) from e
        except TypeError as e:
            error_message = (
                "The model did not return a text response that could be processed. "
                "This can happen if the `max_tokens` parameter is too low, cutting off the response. "
                f"Details: {e}\n"
                f"Model Response: {response_text}"
            )
            raise TypeError(error_message) from e
        except Exception as e:
            # Catching other potential errors, such as API connection issues
            error_message = f"An unexpected error occurred: {e}"
            raise RuntimeError(error_message) from e

    def generate_csv_schema_report(
        self,
        model,
        prompt=None,
        system_instructions=None,
        return_json=False
    ):
        """Generate a CSV schema from a string using the Google GenAI API.

        Args:
            model (str): The name of the model to use.
            prompt (str, optional): The prompt to send to the model. If not provided, a default prompt will be used.
            system_instructions (str, optional): System instructions to guide the model's response.
            return_json (bool, optional): Whether to return the response as a JSON string or leave as a dict.

        Returns:
            dict: The generated schema from the CSV string.
        """
        if not prompt:
            csv_string = CSVStringSample(self.filepath).csv_string_sample_by_quality
            prompt = prompts.CSV_SCHEMA_PROMPT.format(csv_sample_string=csv_string)
        if not system_instructions:
            system_instructions = prompts.CSV_SCHEMA_SYSTEM_INSTRUCTIONS

        csv_schema_report = self._get_ai_response(model, prompt, system_instructions)

        if return_json:
            return json.dumps(csv_schema_report, indent=4)

        return csv_schema_report



"""Module to create engines for interfacing with different large language models and providers."""

# standard library imports
from abc import ABC, abstractmethod
from dataclasses import dataclass

# third-party imports
from google import genai
from google.genai import types

@dataclass
class AIEngineProperties:
    """Base properties for AI Engines."""
    valid_engines: tuple = ('google')

class BaseAIEngine(ABC):
    """Abstract base class for AI Engines."""

    def __init__(self, api_key=None):
        """Initialize the AI provider."""
        self.api_key = api_key

    @abstractmethod
    def generate_content(
        self,
        prompt,
        max_tokens,
        temperature,
        **kwargs
    ):
        """Generate text from a prompt.

        Args:
            prompt: The prompt to send to the model
            max_tokens: Maximum tokens to generate
            temperature: Sampling temperature
            **kwargs: Additional provider-specific parameters

        Returns:
            Generated text
        """
        pass

    @abstractmethod
    def generate_embeddings(
        self,
        text,
        **kwargs
    ):
        """Generate embeddings for text.

        Args:
            text: Text to embed
            **kwargs: Additional provider-specific parameters

        Returns:
            Embedding vector
        """
        pass

class GoogleAIEngine(BaseAIEngine):
    """Class to interact with the Google GenAI API."""

    THINKING_BUDGET = -1  # Default thinking budget for the model. -1 means "auto".
    DEFAULT_RESPONSE_JSON_MIME_TYPE = "application/json"  # Default response MIME type
    MAX_OUTPUT_TOKENS = 8192  # Maximum output tokens for the model
    DEFAULT_TEMPERATURE = 0.5  # Default temperature for the model
    DEFAULT_TOP_P = 1  # Default top_p for the model
    DEFAULT_SEED = 0  # Default seed for the model
    DEFAULT_SYSTEM_INSTRUCTIONS = ""  # Default system instructions for the model

    def __init__(
            self,
            api_key=None,
            vertexai=False,
            gcp_project=None,
            gcp_location=None,
            prompt=None,
            max_tokens=MAX_OUTPUT_TOKENS,
            temperature=DEFAULT_TEMPERATURE,
            top_p=DEFAULT_TOP_P,
            seed=DEFAULT_SEED,
            safety_settings=None,
            thinking_budget=THINKING_BUDGET,
            response_type=DEFAULT_RESPONSE_JSON_MIME_TYPE,
            ground_google_search=False
            ):

        """Initialize the Google AI provider."""
        super().__init__(api_key)
        self.vertexai = vertexai
        if not self.api_key and not self.vertexai:
            raise ValueError("Either api_key or vertexai must be provided.")
        self.gcp_project = gcp_project
        self.gcp_location = gcp_location
        self.prompt = prompt
        self.max_tokens = max_tokens
        self.temperature = temperature
        self.top_p = top_p
        self.seed = seed
        self.safety_settings = safety_settings if safety_settings is not None else self._safety_settings()
        self.thinking_budget = thinking_budget
        self.response_type = response_type
        self.ground_google_search = ground_google_search
        if not self.api_key and self.vertexai and (not self.gcp_project or not self.gcp_location):
            raise ValueError("You must provide gcp_project and gcp_location when using Vertex AI.")

    def _client(self):
        """Create and return a GenAI client."""
        if self.vertexai:
            return genai.Client(
                vertexai=self.vertexai,
                project=self.gcp_project,
                location=self.gcp_location
            )
        return genai.Client(api_key=self.api_key)

    def _contents(self, prompt):
        """Create content for the model based on the provided prompt.

        Args:
            prompt (str): The prompt to send to the model.
        Returns:
            list: A list of Content objects with the user role and parts.
        """
        return [
            types.Content(
                role="user",
                parts=[
                    types.Part.from_text(text=prompt)],
            ),
        ]

    def _safety_settings(self):
        """Create safety settings for the model."""
        return [
            types.SafetySetting(
                category="HARM_CATEGORY_HATE_SPEECH",
                threshold="OFF"
            ),
            types.SafetySetting(
                category="HARM_CATEGORY_DANGEROUS_CONTENT",
                threshold="OFF"
            ),
            types.SafetySetting(
                category="HARM_CATEGORY_SEXUALLY_EXPLICIT",
                threshold="OFF"
            ),
            types.SafetySetting(
                category="HARM_CATEGORY_HARASSMENT",
                threshold="OFF"
            )
        ]

    def _ground_in_google_search(self):
        """Ground the model in Google Search."""
        return [
                 types.Tool(google_search=types.GoogleSearch()),
            ]

    def _content_config(self, system_instruction=None):
        """Content configuration for API calls to the LLM."""
        if not system_instruction:
            system_instruction = self.DEFAULT_SYSTEM_INSTRUCTIONS

        if self.ground_google_search:
            config = types.GenerateContentConfig(
                temperature = self.temperature,
                top_p = self.top_p,
                seed = self.seed,
                max_output_tokens = self.max_tokens,
                safety_settings = self.safety_settings,
                system_instruction=[types.Part.from_text(text=system_instruction)],
                thinking_config=types.ThinkingConfig(
                    thinking_budget=self.thinking_budget,
                ),
                response_mime_type = self.response_type,
                tools = self._ground_in_google_search()
            )
        else:
            config = types.GenerateContentConfig(
                temperature = self.temperature,
                top_p = self.top_p,
                seed = self.seed,
                max_output_tokens = self.max_tokens,
                safety_settings = self.safety_settings,
                system_instruction=[types.Part.from_text(text=system_instruction)],
                thinking_config=types.ThinkingConfig(
                    thinking_budget=self.thinking_budget,
                ),
                response_mime_type = self.response_type,
            )
        return config

    def generate_content(self, model, prompt, system_instruction=None):
        """Generate content using the Google GenAI API.

        Args:
            model (str): The name of the model to use.
            prompt (str): The prompt to send to the model.
            system_instruction (str, optional): System instructions to guide the model's response.
        Returns:
            str: The generated response from the model.
        """
        contents = self._contents(prompt)
        generate_content_config = self._content_config(system_instruction)

        response = self._client().models.generate_content(
            model=model,
            contents=contents,
            config=generate_content_config,
        )

        return response.text

    def generate_embeddings(self):
        raise NotImplementedError("Embedding generation is not yet implemented for the Google provider.")



"""Initializes the ai module of the datagrunt package."""

from datagrunt.core.ai.engines import AIEngineProperties, GoogleAIEngine
from datagrunt.core.ai.factories import AIEngineFactory
from datagrunt.core.ai.prompts import (
    CSV_SCHEMA_SYSTEM_INSTRUCTIONS,
    CSV_SCHEMA_PROMPT,
    SUGGEST_DATA_TRANSFORMATIONS,
    GENERATE_SQL_QUERY
)

__all__ = [
    'AIEngineProperties',
    'GoogleAIEngine',
    'AIEngineFactory',
    'CSV_SCHEMA_SYSTEM_INSTRUCTIONS',
    'CSV_SCHEMA_PROMPT',
    'SUGGEST_DATA_TRANSFORMATIONS',
    'GENERATE_SQL_QUERY'
]



"""Initializes the core module of the datagrunt package."""

from datagrunt.core.ai import (
    AIEngineFactory,
    AIEngineProperties,
    GoogleAIEngine,
    CSV_SCHEMA_PROMPT,
    CSV_SCHEMA_SYSTEM_INSTRUCTIONS,
    GENERATE_SQL_QUERY,
    SUGGEST_DATA_TRANSFORMATIONS,
    prompts
)
from datagrunt.core.csv_io import (
    CSVEngineFactory,
    CSVReaderDuckDBEngine,
    CSVReaderPolarsEngine,
    CSVWriterDuckDBEngine,
    CSVWriterPolarsEngine,
    CSVEngineProperties,
    CSVComponents,
    CSVColumnNameNormalizer,
    CSVColumns,
    CSVDelimiter,
    CSVDialect,
    CSVRows,
    CSVStringSample
)
from datagrunt.core.databases import DuckDBDatabase, DuckDBQueries
from datagrunt.core.file_io import FileProperties

__all__ = [
    # AI
    'AIEngineFactory',
    'AIEngineProperties',
    'GoogleAIEngine',
    'CSV_SCHEMA_PROMPT',
    'CSV_SCHEMA_SYSTEM_INSTRUCTIONS',
    'GENERATE_SQL_QUERY',
    'SUGGEST_DATA_TRANSFORMATIONS',
    'prompts',
    # CSV IO
    'CSVEngineFactory',
    'CSVReaderDuckDBEngine',
    'CSVReaderPolarsEngine',
    'CSVWriterDuckDBEngine',
    'CSVWriterPolarsEngine',
    'CSVEngineProperties',
    'EngineProperties',
    'CSVComponents',
    'CSVColumnNameNormalizer',
    'CSVColumns',
    'CSVDelimiter',
    'CSVDialect',
    'CSVRows',
    'CSVStringSample',
    # Databases
    'DuckDBDatabase',
    'DuckDBQueries',
    # File IO
    'FileProperties',
]


"""This module contains prompts for the AI models."""


CSV_SCHEMA_SYSTEM_INSTRUCTIONS = """
You are a data engineering agent.
Your job is as follows:
    * evaluate data samples to identify columns
    * determine data types
    * classify dimensions from measures
    * determine if the data has a column header or not
    * determine if the data is structured or unstructured
    * determine if the data has a delimiter and to identify that delimiter.
Categorize these finding into their own section of the response format.
"""

CSV_SCHEMA_PROMPT = """
For the given data string identify the columns, return a list of columns, the column data types, and whether or not they should be categorized as a dimension or a measure.
Also return a normalized version of the data column name in all lower case separated by underscores. If the column name starts with a number move the number to the end of the column name.
Anything that could be categorized as a special number could lose a leading zero if converted from text to a numeric type, be sure in those cases to classify the column as a dimension and make it a string type.
Here are some examples of special values that could be numeric typeable but should be classified as dimensions and string types:

<special values>
    * ZIP Codes / Postal Codes
    * Product SKUs / Item Numbers
    * Employee IDs / Customer IDs / User IDs
    * Phone Numbers (if stored numerically)
    * Bank Account Numbers
    * Serial Numbers
    * Course Codes / Class IDs
    * Lot Numbers / Batch Numbers
    * Dates / Times (if represented as a single number without separators, e.g., MMDD)
    * Social Security Numbers (SSNs)
    * Transaction Codes
    * Any number with a leading zero in the original data

If unsure if a data value should be categorized as a dimension or measure, for safety sake, categorize as a dimension and string type.
Return a map of original column names to normalized column names.

<example measures>
    * Sales Amount
    * Quantity Sold
    * Revenue
    * Profit Margin
    * Temperature
    * Distance

If `has_column_header` is false, then recommend column header names based on the the data that are nicely formatted in lowercase and underscore separated.
After generating a list of recommended columns, add another element grouping those recommendations into dimension and measures.
Be sure to reserve measures only for quantifiable values and not anything that could be categorized as a `special value`.

Return a response format like this:

<response format>

{{
 "has_column_header": true or false ,
 "is_structured": true or false ,
 "has_delimiter": true or false,
 "delimiter": "delimiter",
 "is_tabular": true or false,
 "encoding": "utf-8", "latin1", "latin2", "utf-16", etc.,
 "total_column_count": 0,
 "total_dimension_column_count": 0,
 "total_measure_column_count": 0,
 "schema": [
  {{
   "name": "column name",
   "normalized_name": "normalized_column_name",
   "data_type": "data_type",
   "classification": "dimension or measure"
  }}
 ],
 "dimensions": ["colmn_name_1", "column_name_2", ...],
 "measures": ["column_name_1", "column_name_2", ...],
"columns_rename_map": {{
    "column_name": "normalized_column_name"
    }}
}}

<data string>
```{csv_sample_string}```
"""

SUGGEST_DATA_TRANSFORMATIONS = """
Given this CSV data sample:

{csv_sample}

The user wants to: {user_goal}

Please suggest specific data transformations that would help achieve this goal.
Provide your suggestions as a numbered list of clear, actionable steps.
Focus on transformations that can be implemented using SQL or dataframe operations.
"""

GENERATE_SQL_QUERY = """
Given a table named '{table_name}' with the following schema:

{schema_description}

Generate a SQL query for this request: {natural_language_query}

Return ONLY the SQL query without any explanation or markdown formatting.
The query should be compatible with DuckDB syntax.
"""
